shuffuled_frog_optimization = """
import numpy as np

def opt_func(x):
    return np.sqrt(np.sum(np.square(x)))

def gen_frogs(frogs, dim, sigma, mu):
    return np.random.normal(mu, sigma, (frogs, dim))

def sort_frog(frogs, mplx_no, opt_func):
    fitness = np.array([opt_func(frog) for frog in frogs])

    sorted_idx = np.argsort(fitness)

    memeplexes = np.array_split(sorted_idx, mplx_no)

    return np.array(memeplexes)

def local_search(frogs, memeplex, opt_func, sigma, mu):

    frog_g = frogs[0]

    best_frog = frogs[int(memeplex[0])]
    worst_frog = frogs[int(memeplex[-1])]

    new_worst_frog = worst_frog + (best_frog - worst_frog) * np.random.rand()

    if opt_func(new_worst_frog) > opt_func(worst_frog):

        new_worst_frog = worst_frog + (frog_g - worst_frog) * np.random.rand()

    if opt_func(new_worst_frog) > opt_func(worst_frog):

        new_worst_frog = gen_frogs(1, len(frog_g), sigma, mu)[0]

    frogs[int(memeplex[-1])] = new_worst_frog

    return frogs

def shuffled_leaping_frog(
    opt_func,
    frogs = 30,
    dim = 3,
    sigma = 1,
    mu = 0,
    mplx_no = 3,
    max_iter = 100,
    solution_iter = 1000
    ):

    frogs = gen_frogs(frogs, dim, sigma, mu)

    memeplexes = sort_frog(frogs, mplx_no, opt_func)

    best_solution = frogs[int(memeplexes[0, 0])]

    for i in range(solution_iter):

        if i % 10 == 0:
            print(f"Iteration {i}: best value: {opt_func(best_solution):.3f}", end=" ")
            print(f"Best position: {np.around(best_solution, 3)}")


        # Shuffle memeplexes
        np.random.shuffle(memeplexes)

        for memeplex_idx,  memeplex in enumerate(memeplexes):
            for j in range(max_iter):
                frogs = local_search(frogs, memeplex, opt_func, sigma, mu)

            memeplexes = sort_frog(frogs, mplx_no, opt_func)

            if opt_func(frogs[int(memeplexes[0, 0])]) < opt_func(best_solution):
                best_solution = frogs[int(memeplexes[0, 0])]


    return best_solution, opt_func(best_solution)

print("Starting shuffled leaping frog algorithm")

best_pos, best_val = shuffled_leaping_frog(
    opt_func,
    100,
    3,
    1,
    0,
    5,
    250,
    500
)

print("Completed shuffled leaping frog algorithm")

print(f"Best position: {np.around(best_pos, 3)}")
print(f"Best value: {best_val:.3f}")
"""


continuous_optimization = """
from numpy.random import randint
from numpy.random import rand


def objective(x):
    return x[0] ** 2.0 + x[1] ** 2.0


def decode(bounds, n_bits, bitstring):
    decoded = list()

    largest = 2 ** n_bits

    for i in range(len(bounds)):
        start, end = i * n_bits, i * n_bits + n_bits 

        substring = bitstring[start: end]

        chars = "".join([str(s) for s in substring])


        integer = int(chars, 2)

        value = bounds[i][0] + (integer / largest) * bounds[i][1] -bounds[i][0]

        decoded.append(value)

    return decoded

def selection(pop, scores, k=3):
    selection_ix = randint(len(pop))

    for ix in randint(0, len(pop), k-1):
        if scores[ix] < scores[selection_ix]:
            selection_ix = ix

    return pop[selection_ix]


def crossover(p1, p2, r_cross):

    c1, c2 = p1.copy(), p2.copy()

    if rand() < r_cross:
        pt = randint(1, len(p1) - 2)

        c1 = p1[:pt] + p2[pt:]
        c2 = p2[:pt] + p1[pt:]

    return [c1, c2]

def mutation(bitstring, r_mut):
    for i in range(len(bitstring)):
        if rand() < r_mut:
            bitstring[i] = 1 - bitstring[i]


def genetic_algorithm(
        objective, bounds, n_bits, n_iter,
        n_pop, r_cross, r_mut):
    pop = [
            randint(0, 2, n_bits*len(bounds)).tolist()
            for _ in range(n_pop)
        ]
    best, best_eval = 0, objective(decode(bounds, n_bits, pop[0]))
    for gen in range(n_iter):
        decoded = [
            decode(bounds, n_bits, p)
            for p in pop
        ]
        scores = [objective(d) for d in decoded]

        for i in range(n_pop):
            if scores[i] < best_eval:

                best, best_eval = pop[i], scores[i]

                print("> iteration %d, new best f(%s) = %f" % (gen, decoded[i], scores[i]))

        selected = [selection(pop, scores) for _ in range(n_pop)]
        children = list()

        for i in range(0, n_pop, 2):
            p1, p2 = selected[i], selected[i+1]
            for c in crossover(p1, p2, r_cross):
                mutation(c, r_mut)      
                children.append(c)

        pop = children

    return [best, best_eval]
bounds = [[-5.0, 5.0], [-5.0, 5.0]]
n_iter = 1000
n_bits = 160
n_pop = 100
r_cross = 0.9
r_mut = 1.0 / (float(n_bits) * len(bounds))


print("Starting genetic algorithm")
best, score = genetic_algorithm(objective, bounds, n_bits, n_iter, n_pop, r_cross, r_mut)
decoded = decode(bounds, n_bits, best)
print("Genetic algorithm complete")
print(f"Best Solution: {decoded}")
print(f"Fitness Score of the best solution: {score:.2f}")
"""


particle_swarn_optimization  = """
import random
import sys
import numpy as np
from scipy.spatial import distance

target = np.random.uniform(-10, 10, 3)
print("Generated Target Position: ", target)

def fitness_sphere(position):
    return distance.euclidean(position, target)


class Particle:
    def __init__(self, fitness, dim, minx, maxx, seed):
        self.position = np.random.uniform(minx, maxx, dim)
        self.velocity = np.random.uniform(minx, maxx, dim)

        self.fitness = fitness(self.position)
        self.best_part_pos = np.copy(self.position)
        self.best_part_val = self.fitness

def pso(fitness, max_iter, n, dim, minx, maxx):

    w = 0.729
    c1 = 1.49445
    c2 = 1.49445

    swarm = [Particle(fitness, dim, minx, maxx, i) for i in range(n)]

    best_swarm_particle = max(
        swarm,
        key = lambda x: x.fitness
    )

    best_swarm_pos = best_swarm_particle.position
    best_swarm_val = best_swarm_particle.fitness

    for t in range(max_iter):

        if t % 10 == 0:
            print(f"Iteration {t}: best value: {best_swarm_val:.3f}", end=" ")
            print(f"Best position: {np.around(best_swarm_pos, 3)}")

        for p in swarm:
            p.velocity = w * p.velocity + \
                c1 * np.random.rand(dim) * (p.best_part_pos - p.position) + \
                c2 * np.random.rand(dim) * (best_swarm_pos - p.position)

            p.velocity = np.clip(p.velocity, minx, maxx)
            p.position = p.position + p.velocity

            p.fitness = fitness(p.position)

            if p.fitness < p.best_part_val:
                p.best_part_pos = np.copy(p.position)
                p.best_part_val = p.fitness

            if p.fitness < best_swarm_val:
                best_swarm_pos = p.position
                best_swarm_val = p.fitness
    return best_swarm_pos, best_swarm_val

dim = target.shape[0]
num_particles = 50
max_iter = 100

print("Starting PSO")
best_pos, best_val = pso(fitness_sphere, max_iter, num_particles, dim, -10., 10.)

print("Completed PSO")
print(f"Best position: {np.around(best_pos, 3)}")
print(f"Best value: {best_val:.3f}")
"""


ant_colony_algorithm = """
from networkx.algorithms.centrality.current_flow_betweenness_subset import edge_current_flow_betweenness_centrality_subset
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from itertools import combinations

def plot_graph(
    g,
    title = " ",
    highlight_edges = [],
    ):

    pos = nx.get_node_attributes(g, "pos")

    plt.figure(figsize=(17, 17))
    plt.title(title)
    nx.draw(g, pos=pos, labels={x: x for x in g.nodes}, width=2)
    weights = nx.get_edge_attributes(g, "weight")
    # Draw edge labels
    nx.draw_networkx_edge_labels(g, pos, edge_labels=weights, label_pos=.4)
    # Highlight edges
    nx.draw_networkx_edges(g, pos, edgelist=highlight_edges, edge_color="r")

    nx.draw_networkx_edge_labels(
        g,
        pos,
        edge_labels={
            e:w
            for e, w in weights.items()
            if e in map(lambda x: tuple(sorted(x)), highlight_edges)
        },
        font_color="r",
        label_pos=.4
    )

    plt.show()

def generate_random_weighted_graph(n, low, high):
    g = nx.generators.complete_graph(n)

    g.add_weighted_edges_from([
        (u, v, np.random.randint(low, high))
        for u, v in g.edges()
    ])

    nx.set_node_attributes(g, nx.spring_layout(g), "pos")

    return g

def zero_divide(a, b):
    return np.divide(a, b, out=np.zeros_like(a), where=b!=0)


class ACOTSP:
    def __init__(self, g, n_ants=100, alpha=1, beta=5, Q=100, rho=.6):
        self.g = g
        self.n_nodes = g.number_of_nodes()

        distances = nx.to_numpy_array(g)
        print(distances)

        self.visibility = zero_divide(np.ones_like(distances), distances)

        self.n_ants = n_ants
        self.alpha = alpha
        self.beta = beta

        self.Q = Q
        self.rho = rho

        self.pheromone = np.ones((self.n_nodes, self.n_nodes))


    def compute_probabilities(self, visited):
        self.prob = self.pheromone ** self.alpha * self.visibility ** self.beta

        self.prob[:, np.array(list(visited))] = 0

        prob_sum = self.prob.sum(-1, keepdims = True)

        self.prob = zero_divide(self.prob, prob_sum)

        return self.prob

    def initialize_ants(self):
        nodes = list(self.g.nodes)

        self.ant_pos = np.random.choice(nodes, self.n_ants)


    def path_length(self, path):
        edge_weights = nx.get_edge_attributes(self.g, "weight")

        return sum(edge_weights[tuple(sorted(edge))] for edge in path)

    def ant_tour(self, k):
        current = self.ant_pos[k]

        visited = [current]
        path = []

        self.compute_probabilities(visited)

        while True:
            prev = int(current)

            current = np.random.choice(self.n_nodes, p = self.prob[current])

            visited.append(current)
            path.append((prev, current))
            if np.all(self.prob[current] == 0):
                break

            self.compute_probabilities(visited)

        self.paths[k] = path

    def update_pheromone(self):
        d_pheromone = np.zeros((self.n_nodes, self.n_nodes, self.n_ants))

        for k, path in enumerate(self.paths):
            if len(path) == self.n_nodes - 1:
                for u, v in path:
                    d_pheromone[u, v, k] = d_pheromone[v, u, k] = self.Q / self.path_length(path)

        d_pheromone = d_pheromone.sum(-1)
        self.pheromone = self.rho * self.pheromone + d_pheromone

    def run(self, n_iter = 1):
        self.initialize_ants()
        self.min_path = None
        for t in range(n_iter):
            self.paths = [[None]] * self.n_ants

            for k in range(self.n_ants):
                self.ant_tour(k)

            self.update_pheromone()

            self.hamiltonian_paths = [
                path
                for path in self.paths
                if len(path) == self.n_nodes - 1
            ]


            if t % 10 == 0:
                print(f"Iteration {t}: Best path: {self.best_path} with length {self.best_path_length}")


    @property
    def best_path_length(self):
        return self.path_length(self.best_path)

    @property
    def best_path(self):
        if not hasattr(self, "hamiltonian_paths") or len(self.hamiltonian_paths) == 0:
            return None
        return min(self.hamiltonian_paths if self.min_path is None else [self.min_path, *self.hamiltonian_paths] , key=self.path_length)



g = generate_random_weighted_graph(4, 10, 50)


acotsp_solver = ACOTSP(g, n_ants=100, alpha=1, beta=5, Q=100, rho=.4)

acotsp_solver.run(100)

plot_graph(g, "Shortest Path found by ACO for TSP", acotsp_solver.best_path)
"""


simulated_annealing = """
from numpy import asarray, exp
from numpy.random import randn, rand, seed
import numpy as np

def objective(x):
    return x[0]**2.0

def simulate_annealing(
        objective,
        bounds,
        n_iterations,
        step_size,
        temp
        ):

    best = np.random.uniform(bounds[:, 0], bounds[:, 1])

    best_eval = objective(best)

    curr, curr_eval = best, best_eval


    for i in range(n_iterations):
        candidate = curr + rand(len(bounds)) * step_size

        candidate_eval = objective(candidate)

        if candidate_eval < best_eval:
            best, best_eval = candidate, candidate_eval
            print("> Iteration %d: f(%s) = %.5f" % (i, best, best_eval))

        diff = candidate_eval - curr_eval
        t = temp / float(i + 1)
        metropolis = exp(-diff/ t)
        if diff < 0 or rand() < metropolis:
            curr, curr_eval = candidate, candidate_eval


    return [best, best_eval]

seed(1)
bounds = asarray([[-5.0, 5.0]])
n_iterations = 1000

step_size = 0.1
temp = 10

print("Started Simulated Annealing Algorithm")
best, score = simulate_annealing(

        objective,
        bounds,
        n_iterations,
        step_size,
        temp
    )
print("\n Simulated Annealing Completed \n")
print(f"")
"""


tsp = """
import numpy as np
import networkx as nx

import matplotlib.pyplot as plt
from collections import defaultdict
from itertools import combinations

def generate_random_weighted_graph(n, low, high):
    g = nx.generators.complete_graph(n)

    g.add_weighted_edges_from([
        (u, v, np.random.randint(low, high))
        for u, v in g.edges()
    ])

    nx.set_node_attributes(g, nx.spring_layout(g), "pos")
    return g


    def plot_graph(g, title="", highlight_edges=[]):
    pos = nx.get_node_attributes(g, "pos")
    plt.figure(figsize=(17, 17))

    plt.title(title)
    nx.draw(g, pos=pos, labels={x:x for x in g.nodes}, width=2)

    weights = nx.get_edge_attributes(g, "weight")

    nx.draw_networkx_edge_labels(g, pos, edge_labels=weights, label_pos=.4)

    nx.draw_networkx_edges(g, pos, edgelist=highlight_edges, edge_color="r", width=3)

    nx.draw_networkx_edge_labels(
        g, 
        pos, 
        edge_labels={
            e: w 
            for e, w in weights.items() 
            if e in map(lambda x: tuple(sorted(x)), highlight_edges)
        }, 
        font_color="r", 
        label_pos=.4
    )
    plt.show()

def roulette_wheel_selection(fitness_func,parents):
    fitness = np.array(list(map(fitness_func,parents)))
    fitness = fitness/fitness.sum()
    n = len(parents)
    return [parents[i] for i in np.random.choice(n,n,p=fitness)]

    
g = generate_random_weighted_graph(7, 1, 20)

plot_graph(g, "Graph for TSP")

def nearest_neighbour_initialization(g, closed_tour=False):
    curr_node = int(np.random.choice(g.nodes))
    path = [curr_node]

    not_visited = set(g.nodes) - {curr_node}

    while not_visited:
        not_visited_neigbors = not_visited & set(g.neighbors(curr_node))
        key = lambda x: g[curr_node][x]["weight"]
        curr_node = min(not_visited_neigbors, key=key)
        path.append(curr_node)
        not_visited.remove(curr_node)

    if closed_tour:
        path.append(path[0])

    return path

np.random.seed(1)
print(nearest_neighbour_initialization(g))
print(nearest_neighbour_initialization(g, True))

from collections import defaultdict

def has_cycle(g):
    try:
        nx.find_cycle(g)
    except nx.NetworkXNoCycle:
        return False
    return True

def get_path_from_edges(edges, closed_tour=False):
    path_graph = nx.Graph(edges)

    curr = min(path_graph, key=path_graph.degree)

    path, visited = [curr], {curr}

    while len(path) < len(path_graph):
        curr = (set(path_graph.neighbors(curr)) - visited).pop()
        visited.add(curr)
        path.append(curr)

    if closed_tour:
        path.append(path[0])

    return path
    
def select_stochastic(a,key,inverse=True):
    a = list(a)
    probs = np.array(list(map(key,a)))
    if inverse:
        probs = 1/np.array(probs)
    probs = probs/probs.sum()
    return type(a[0])(a[np.random.choice(len(a),p=probs)])

def shortest_edge_initialization(g, closed_tour=False, stochastic=True):
    edge_list = set(g.edges)

    times_visited = defaultdict(int)
    tour = set()
    max_tour_len = len(g) if closed_tour else len(g) - 1
    key = nx.get_edge_attributes(g, "weight").get

    while len(tour) < max_tour_len:
        if not stochastic:
            u, v = min(edge_list, key=key)
        else:
            u, v = select_stochastic(edge_list, key=key)
        times_visited[u] += 1
        times_visited[v] += 1

        tour.add((u, v))

        edge_list.remove((u,v))
        for u, v in set(edge_list):
            if (
                (has_cycle(nx.Graph(tour|{(u, v)})) and len(tour) != len(g) - 1)
                or times_visited[u] == 2 or times_visited[v] == 2
            ):
                edge_list.remove((u,v))

    return get_path_from_edges(tour, closed_tour=closed_tour)

np.random.seed(1)
print(shortest_edge_initialization(g))
print(shortest_edge_initialization(g, True))
    

def make_valid_tour(p, nodes):
    unvisited = set(nodes) - set(p)
    indices = defaultdict(list)

    for i in range(len(p)):
        indices[p[i]].append(i)
    visited_twice = { node for node in indices if len(indices[node]) == 2 }
    for node in visited_twice:
        change_index = np.random.choice(indices[node])
        p[change_index] = unvisited.pop()

    return p

def partially_matched_crossover(p1, p2):
    pt = np.random.randint(1, len(p1)-1)
    c1 = p1[:pt] + p2[pt:]
    c2 = p2[:pt] + p1[pt:]


    nodes = set(p1)
    return make_valid_tour(c1, nodes), make_valid_tour(c2, nodes)

inv_path_length = lambda x: 1/nx.path_weight(g,x,"weight")


n_population = 8
population = [
    shortest_edge_initialization(g, closed_tour=False, stochastic=True)
    for _ in range(n_population)
]  

selected_population = roulette_wheel_selection(inv_path_length, population)
parents = selected_population[:2]

print(parents)

print(partially_matched_crossover(*parents))

def order_crossover(p1, p2):
    start = np.random.randint(0, len(p1) - 1)
    end = np.random.randint(start + 1, len(p1) if start != 0 else len(p1) - 1)

    def fill_blanks(p1, p2, s, e):
        unvisited_nodes = p2.copy()

        for node in p1[s:e]:
            unvisited_nodes.remove(node)

        c = p1.copy()
        for i in range(len(p1)):
            if i < s or i >=e:
                c[i] = unvisited_nodes.pop(0)
        return c

    c1 = fill_blanks(p1, p2, start, end)
    c2 = fill_blanks(p2, p1, start, end)

    return c1, c2
   
n_population = 8
population = [shortest_edge_initialization(g, closed_tour=False) for _ in range(n_population)]

selected_population = roulette_wheel_selection(inv_path_length, population)
parents = selected_population[:2]

print(parents)
print(order_crossover(*parents))

def inversion_mutation(p):
    start = np.random.randint(1, len(p) - 1)
    end = np.random.randint(start+1, len(p) + 1)
    print(start, end)
    c = p.copy()

    c[start:end] = reversed(c[start:end])
    return c

n_population = 8
population = [
    shortest_edge_initialization(g, closed_tour=False)
    for _ in range(n_population)
]

subject = population[0]
print(subject, inversion_mutation(subject), sep="\n")

def insertion_mutation(p):
    i = np.random.randint(1, len(p))
    k = np.random.randint(0, len(p) - 1)

    print(i, k)
    c = p.copy()
    c.insert(k, c.pop(i))
    return c

n_population = 8 
population = [
    shortest_edge_initialization(g, closed_tour=False)
    for _ in range(n_population)
]

subject = population[0]

print(subject, insertion_mutation(subject), sep="\n")
"""


from setuptools import setup, find_packages

setup(
    name='dharun',
    version='1.2',
    packages=find_packages(),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Dharun',
    author_email='leodas@gmail.com',
    description='Some Optimization Problems for Machine Learning',
    url='https://github.com/KeerthiVasan-ai/optimization-techniques', 
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License', 
        'Operating System :: OS Independent',
    ],
)



gray_wolf_optimization = """
import random
import numpy as np

from scipy.spatial import distance


target = np.random.uniform(-100, 100, 3)
print("Generated Target Position: ", target)

def fitness_sphere(position):
    return distance.euclidean(position, target)

class Wolf:
    def __init__(self, fitness, dim, minx, maxx, seed=0):

        np.random.seed(seed)
        self.position = np.random.uniform(minx, maxx, dim)

        self.fitness = fitness(self.position)

def gwo(fitness, max_iter, n, dim, minx, maxx):

    population = [
        Wolf(fitness, dim, minx, maxx, i) for i in range(n)
    ]

    population = sorted(population, key=lambda x: x.fitness)

    alpha, beta, gamma = population[:3].copy()

    for t in range(max_iter):

        wolf_pos = np.array([wolf.position for wolf in population[:3]])

        if t % 10 == 0:
            print(f"Iteration {t}: best value: {alpha.fitness:.3f}", end=" ")
            print(f"Best position: {np.around(alpha.position, 3)}")

        a = 2 * (1 - t / max_iter)

        for i, wolf in enumerate(population):
            A = a * np.random.uniform(-1, 1, (3, 1))

            C = 2 * np.random.uniform(0, 1, (3, 1))

            X = wolf_pos - A * np.abs(C * wolf_pos - wolf.position)

            Xnew = X.mean(axis=0)

            if fitness(Xnew) < wolf.fitness:
                wolf.position = Xnew
                wolf.fitness = fitness(Xnew)

        population = sorted(population, key=lambda x: x.fitness)
        alpha, beta, gamma = population[:3].copy()

    return alpha.position, alpha.fitness


dim = target.shape[0]
n = 100
max_iter = 100

print("Starting GWO")

best_pos, best_val = gwo(fitness_sphere, max_iter, n, dim, -100., 100.)

print("Completed GWO")

print(f"Best position: {np.around(best_pos, 3)}")

print(f"Best value: {best_val:.3f}")
"""


tabu_search = """
import pandas as pd
import numpy as np
from itertools import combinations

class TS:
    def __init__(self, path, seed, tabu_tenure):
        self.path = path
        self.seed = seed
        self.tabu_tenure = tabu_tenure

        self.instance_dict = self.input_data()
        self.initial_solution = self.get_initial_solution()

        self.tabu_str, self.best_solution, self.best_obj_val = self.tabu_search()

    def input_data(self):
        return pd.read_csv(self.path, names=[
            "job", "weight", "processing_time","due_date"
        ], index_col=0).to_dict("index")

    def get_tabu_structure(self):
        return {
            swap: {
                "tabu_time": 0,
                "move_val": 0
            }

            for swap in combinations(self.instance_dict.keys(), 2)
        }

    def get_initial_solution(self):
        arr = np.array(list(self.instance_dict.keys()))
        np.random.shuffle(arr)
        return arr

    def obj_func(self, solution):
        dict = self.instance_dict
        t = 0

        obj_val = 0

        for job in solution:
            c_i = t + dict[job]["processing_time"]
            d_i = dict[job]["due_date"]
            t_i = max(0, c_i - d_i)
            w_i = dict[job]["weight"]

            obj_val += w_i * t_i
            t = c_i
        return obj_val

    def swap_move(self, solution, i, j):
        new_solution = np.copy(solution)
        i = np.where(new_solution == i)[0][0]
        j = np.where(new_solution == j)[0][0]
        new_solution[[i, j]] = new_solution[[j, i]]
        return new_solution

    def tabu_search(self):
        tenure = self.tabu_tenure

        tabu_str = self.get_tabu_structure()
        best_solution = np.copy(self.initial_solution)
        best_obj_val = self.obj_func(best_solution)

        current_solution = np.copy(self.initial_solution)
        current_obj_val = best_obj_val

        iter = 1
        terminate = 0
        while terminate  < 100:

            if iter < 10:
                print(f"Iteration {iter}: best value: {best_obj_val:.3f}")

            for move in tabu_str:
                candidate_solution = self.swap_move(current_solution, *move)
                candidate_obj_val = self.obj_func(candidate_solution)

                tabu_str[move]["move_val"]  = candidate_obj_val

            while True:
                best_move = min(
                    tabu_str,
                    key = lambda x: tabu_str[x]["move_val"]
                )

                move_val = tabu_str[best_move]["move_val"]
                tabu_time = tabu_str[best_move]["tabu_time"]

                if tabu_time < iter:

                    current_solution = self.swap_move(current_solution, *best_move)
                    current_obj_val = self.obj_func(current_solution)

                    if move_val < best_obj_val:
                        best_solution = np.copy(current_solution)
                        best_obj_val = move_val
                        terminate = 0
                    else :
                        terminate += 1

                    tabu_str[best_move]["tabu_time"] = iter + tenure
                    iter += 1
                    break
                else:
                    if move_val < best_obj_val:
                        current_solution = self.swap_move(current_solution, *best_move)
                        current_obj_val = self.obj_func(current_solution)
                        terminate = 0
                        iter += 1
                        break
                    else:
                        tabu_str[best_move]["move_val"] = np.inf
                        continue

        print("Tabu search terminated")
        print("Performed Iterations:", iter, "Best Solution:", best_solution, "Best Value: ", best_obj_val, sep="\n")
        return tabu_str, best_solution, best_obj_val

print("Tabu search started")

ts = TS("tabu_jobs.csv", 42, 5)
"""